# -*- coding: utf-8 -*-

# Copyright (c) 2009, 2010 Efrim <efrim@azylum.org> {{{
#
#  This file is part of MPD_sima
#
#  MPD_sima is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public
#  License along with this program.
#  If not, see <http://www.gnu.org/licenses/>.
#
#
#  }}}

"""
Deal with configuration and data files.
Parse configuration file and set defaults for missing options.

DATA FORMAT:
    * USER_DB
    user_db is a dict of [tuples: (artist, match)…]
    user_db['Venetian Snares'] = [('Autechre', 99), ('Aphex Twin', 80)]

    * PLAY HISTORY
    play history is a dequeu list of Track object
"""

# DOC {{{
#    http://mail.python.org/pipermail/python-list/2004-February/248674.html
# }}}

# IMPORTS {{{
from __future__ import with_statement
import ConfigParser
import pickle
import sys

from collections import deque
from datetime import date
from ConfigParser import (MissingSectionHeaderError, ParsingError)
from os import (makedirs, environ, stat, chmod)
from os.path import (join, isdir, isfile, expanduser, getsize)
from stat import (S_IMODE, ST_MODE, S_IRWXO, S_IRWXG)
#}}}

#DEFAULTS{{{
DIRNAME = 'mpd_sima'
CONF_FILE = 'mpd_sima.cfg'
USR_DB = 'user_db.cfg'

# Mandatory user options
M_SECTIONS = ['MPD']
# Optional user options
O_SECTIONS = ['sima', 'log']
O_SECTIONS.extend(M_SECTIONS)

DEFAULT_CONF = {
        'MPD': {
            'host': "localhost",
            'port': "6600",
            'password': "false"},
        'sima': {
            'similarity': "22",
            'top_tracks': "false",
            'user_db': "false",
            'cache': "false",
            'history': "false",
            'history_length': "165",
            'queue_length': "1",
            'track_to_add': "1",
            'main_loop_time': "4"},
        'log': {
            'verbosity': "info"}}

#}}}


class CacheError(Exception):#{{{
    """Exception"""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)#}}}


#CONFIG FONCTIONS{{{


def load_usr_cache(cache_file):#{{{
    """Add User Cache to regular cache"""
    config = ConfigParser.SafeConfigParser()

    def get_default(match):
        """Set a default of 99% if None for match."""
        if not match:
            return 99
        else:
            return float(match)
    try:
        config.read(cache_file)
    except (MissingSectionHeaderError, ParsingError), err_conf:
        raise CacheError(err_conf)

    usr_db = {}
    # Convert config to dict object to match CACHE type
    for artist in config.sections():
        usr_db[artist] = [(art[0], get_default(art[1])) for art in config.items(artist)]

    return usr_db#}}}


class ConfMan(object):#CONFIG MANAGER CLASS{{{
    """
    Configuration manager.
    Default configuration is stored in DEFAULT_CONF dictionnary.
    Then init_config() is run and retrieve configuration from defaults if not
    set in conf files.
    These settings are then updated with command line options with
    supersedes_config_with_cmd_line_options().
    Order of priority for the origin of an option is then (lowest to highest):
        * DEFAULT_CONF
        * configuration file (override previous)
        * command line options (override previous)
    """

    def __init__(self, logger, options=None):#{{{
        # options settings priority:
        # defauts < conf. file < command line
        self.conf_file = None
        self.config = None
        self.defaults = dict(DEFAULT_CONF)
        self.startopt = options
        ## play history
        self.history_file = None
        self.play_history = deque()
        ## last.fm request cache
        self.similar_cache_file = None
        self.similar_cache = dict()
        self.tops_cache = dict()
        self.similar_cache_stamp = dict()
        ## User db
        self.user_db_file = None
        self.user_db = dict()

        self.log = logger
        ## INIT CALLS
        self.init_config()
        self.control_conf()
        self.supersedes_config_with_cmd_line_options()
        self.load_cache()
        self.load_user_db()
        self.load_hist()
        #}}}

    def cache_man(self):#{{{
        """Create/update cache.
        """
        cache = self.similar_cache
        stamp = self.similar_cache_stamp
        self.log.debug(u'Do I save cache?')
        if self.config.getboolean('sima', 'cache') and cache:
            self.write_cache()
            if stamp and 'DATE' in stamp:
                delta = date.today() - stamp['DATE']
                self.log.info(u'Cache is %d days old.' % delta.days)
        else:
            self.log.debug(u'No')#}}}

    def write_cache(self):#{{{
        """
        Write cache to file.
        TODO: controls whether or not cache_file encoding matters here.
        """
        cache_file = self.similar_cache_file
        cache = self.similar_cache
        stamp = self.similar_cache_stamp
        if isfile(cache_file):
            with open(cache_file, 'wb') as cache_fd:
                self.log.info(u'Updating cache with latest requests: %s' %
                              cache_file)
                pickle.dump(cache, cache_fd, -1)
                pickle.dump(stamp, cache_fd, -1)
        elif not isfile(cache_file):
            with open(cache_file, 'wb') as cache_fd:
                stamp = {}
                stamp['DATE'] = date.today()
                self.log.info(u'Creating cache file to %s' % cache_file)
                pickle.dump(cache, cache_fd, -1)
                pickle.dump(stamp, cache_fd, -1)#}}}

    def load_cache(self):#{{{
        """Load cache from file."""
        cache_file = self.similar_cache_file
        if self.config.getboolean('sima', 'cache'):
            self.log.info(u'Loading cache from:  %s' % cache_file)
            if isfile(cache_file):
                with open(cache_file, 'rb') as cache_fd:
                    self.similar_cache = pickle.load(cache_fd)
                    self.similar_cache_stamp = pickle.load(cache_fd)
            if self.similar_cache:
                ## UPGRADE V0.3.0=>0.5.0
                for k in self.similar_cache:
                    if not isinstance(self.similar_cache[k], deque):
                        self.log.error(u'Please purge similar cache. Format has changed in V0.5')
                        sys.exit(2)
                self.log.info(u'%d artists in cache.' % len(self.similar_cache))
        #}}}

    def load_user_db(self):#{{{
        """Load user cache"""
        #
        if self.config.getboolean('sima', 'user_db'):
            self.log.info(u'Loading user DB from: %s' % self.user_db_file)
            if not isfile(self.user_db_file):
                self.log.warning(u'File does not exist…')
                return
            try:
                self.user_db = load_usr_cache(self.user_db_file)
                #print [art[0] for art in self.user_db.items()]
                self.log.info(u'%d artists in user DB' % len(self.user_db))
            except CacheError, cache_err:
                self.log.warning(cache_err)
        #}}}

    def history_man(self):#{{{
        """
        Save play history
        """
        if self.config.getboolean('sima', 'history'):
            self.write_hist()#}}}

    def write_hist(self):#{{{
        """
        Write history to file.
        """
        hist_file = self.history_file
        with open(hist_file, 'wb') as hist_fd:
            self.log.info(u'Saving history to %s' % hist_file)
            pickle.dump(self.play_history, hist_fd, -1)
            #}}}

    def load_hist(self):#{{{
        """
        Load history.
        """
        if not self.config.getboolean('sima', 'history'):
            self.log.info(u'Does not load history, will not save session.')
            return False
        hist_file = self.history_file
        if isfile(hist_file) and getsize(hist_file) != 0:
            with open(hist_file, 'rb') as hist_fd:
                self.log.info(u'Loading history from: %s' % hist_file)
                self.play_history = pickle.load(hist_fd)
                self.log.info(u'%d artists in history' %
                        len(self.play_history))
            ## UPGRADE V0.3.0=>0.5.0
            if not isinstance(self.play_history, deque):
                self.log.error(u'Please purge play history. Format has changed in v0.5')
                sys.exit(2)
        elif not isfile(hist_file):
            self.log.info(u'Will create history file in %s' % hist_file)
        #}}}

    def add_hist(self, track):#{{{
        """
        Add track object to history
        """
        history_len = self.config.getint('sima', 'history_length')
        while len(self.play_history) > history_len:
            ## DEBUG
            ## PARAM
            ## TODO: add history_length parameter
            purge_t = self.play_history.popleft()
            self.log.debug(u'Removing from history: %s - %s' %
                          (purge_t.get_artist(), purge_t.get_title()))
        self.log.debug(u'history: %d' % len(self.play_history))
        self.play_history.append(track)#}}}

#    def dict2conf(self, conf_dict):#{{{
#       """
#      format conf file lines to be written.
#      """
#      a_str = "[%s]\n" % conf_dict
#      for key, val in self.defaults[conf_dict].iteritems():
#          a_str = a_str + "%s = %s\n" % (key, val)
#      a_str = a_str + "\n"
#      return a_str#}}}

    def control_mod(self):#{{{
        """
        Controls conf file permissions.
        """
        mode = S_IMODE(stat(self.conf_file)[ST_MODE])
        self.log.debug(u'file permision is: %o' % mode)
        if mode & S_IRWXO or mode & S_IRWXG:
            self.log.warning(u'File is readable by "other" and/or' +
                             u' "group" (actual permission %o octal).' %
                             mode)
            self.log.warning(u'Consider setting permissions' +
                             u' to 600 octal.')
        #}}}

    def supersedes_config_with_cmd_line_options(self):#{{{
        """Updates defaults settings with command line options"""
        for sec in self.config.sections():
            for opt in self.config.options(sec):
                if opt in self.startopt.keys():
                    self.config.set(sec, opt, str(self.startopt.get(opt)))
                    #print opt
                    #pass
            #}}}

#   def config_write_default(self):#{{{
#       """Write a default configuration file."""
#       ##DEBUG MESS
#       self.log.info(u'No configuration file available, creating one…')
#       self.log.info(u'Configuration file: "%s"' % self.conf_file)
#       #self.log.warning('You may need to tune your MPD host settings…')
#       with open(self.conf_file, 'wu') as fd_conf:
#           for section in M_SECTIONS:
#               # TODO add a try, what if not writable destination
#               fd_conf.write(self.dict2conf(section))
#       chmod(self.conf_file, 0600)
#       #}}}

    def control_conf(self):#{{{
        """Get through options/values and set defaults if not in conf file."""
        # Setting default if not specified
        for section in O_SECTIONS:
            if section not in self.config.sections():
                self.log.debug(u'[%s] NOT in conf file' % section)
                self.config.add_section(section)
                for option in self.defaults[section]:
                    self.config.set(section,
                            option,
                            self.defaults[section][option])
                    self.log.debug(
                            u'Setting option with default value: %s = %s' %
                            (option, self.defaults[section][option]))
            elif section in self.config.sections():
                self.log.debug(u'[%s] present in conf file' % section)
                for option in self.defaults[section]:
                    if self.config.has_option(section, option):
                        self.log.debug(u'option "%s" set to "%s" in conf. file' %
                                      (option, self.config.get(section, option)))
                        #pass
                    else:
                        self.log.debug(
                                u'Option "%s" missing in section "%s"' %
                                (option, section))
                        self.log.debug(u'=> setting default "%s" (may not suit you…)' %
                                       self.defaults[section][option])
                        self.config.set(section, option,
                                        self.defaults[section][option])#}}}

    def init_config(self):#{{{
        """
        Use XDG directory standard if exists
        else use "HOME/(.config|.local/share)/mpd_sima/"
        http://standards.freedesktop.org/basedir-spec/basedir-spec-0.6.html

        Set self.config with user preferences from command line.
        Set self.config with user preferences from conf. file if exists,
        overriding command line.
        """

        homedir = expanduser('~')

        if environ.get('XDG_DATA_HOME'):
            data_dir = join(environ.get('XDG_DATA_HOME'), DIRNAME)
        else:
            data_dir = join(homedir, '.local', 'share', DIRNAME)

        if not isdir(data_dir):
            makedirs(data_dir)
            chmod(data_dir, 0700)

        if environ.get('XDG_CONFIG_HOME'):
            conf_dir = join(environ.get('XDG_CONFIG_HOME'), DIRNAME)
        else:
            conf_dir = join(homedir, '.config', DIRNAME)

        if not isdir(conf_dir):
            makedirs(conf_dir)
            chmod(conf_dir, 0700)

        self.history_file = join(data_dir, 'history.pkl')
        self.similar_cache_file = join(data_dir, 'similar_cache.pkl')

        self.conf_file = join(conf_dir, CONF_FILE)
        self.user_db_file = join(conf_dir, USR_DB)

        config = ConfigParser.SafeConfigParser()

        # Load conf file
        #if not isfile(self.conf_file):
        #    self.config_write_default()

        # If no conf file present, uses defaults
        if not isfile(self.conf_file):
            self.config = config
            return

        self.log.info(u'Loading configuration from:  %s' % self.conf_file)
        self.control_mod()

        try:
            config.read(self.conf_file)
        except (MissingSectionHeaderError, ParsingError), err_conf:
            self.log.error(u'error loading %s:\n%s' %
                           (self.conf_file, err_conf))
            sys.exit(1)

        self.config = config
        ##}}}

#}}}

if __name__ == '__main__':#{{{
    pass
#}}}

# VIM MODLINE
# vim: ai ts=4 sw=4 sts=4 expandtab
